/*
        .88b  d88.  .d8b.  d8888b.     d8888b.  .d8b.  d888888b  .d8b.
        88'YbdP`88 d8' `8b 88  `8D     88  `8D d8' `8b `~~88~~' d8' `8b
        88  88  88 88ooo88 88oodD'     88   88 88ooo88    88    88ooo88
        88  88  88 88~~~88 88~~~       88   88 88~~~88    88    88~~~88
        88  88  88 88   88 88          88  .8D 88   88    88    88   88
        YP  YP  YP YP   YP 88          Y8888D' YP   YP    YP    YP   YP


  d88888b db    db d888888b d8888b.  .d8b.   .o88b. d888888b  .d88b.  d8888b.
  88'     `8b  d8' `~~88~~' 88  `8D d8' `8b d8P  Y8 `~~88~~' .8P  Y8. 88  `8D
  88ooooo  `8bd8'     88    88oobY' 88ooo88 8P         88    88    88 88oobY'
  88~~~~~  .dPYb.     88    88`8b   88~~~88 8b         88    88    88 88`8b
  88.     .8P  Y8.    88    88 `88. 88   88 Y8b  d8    88    `8b  d8' 88 `88.
  Y88888P YP    YP    YP    88   YD YP   YP  `Y88P'    YP     `Y88P'  88   YD

  Copyright 2012 David "DarkCampainger" Braun

  Released under the GNU Lesser General Public License:
    <http://www.gnu.org/licenses/lgpl.html>

  Version 1.0 - July 3, 2012

  Please see 'Demo.dm' for a demo of example usage
  Please see 'Reference.dm' for the library reference
  The library itself is not recommended as reading material

  ---------------------------------------------------------------------------
    -      -      -      -      -      -      -      -      -      -      -
*/


// This library requires at least 495 for new text parsing features
#if DM_VERSION < 495
#error This library requires at least version 495, please update!
#endif

// This will only be defined if the library is included (see demo)
#define MDE_LIBRARY_INCLUDED 1

// This object will generate the mapping data for specific paths
MapDataExtractor

	// List of variables to save (associated default values)
	var/savedVariables = list( "icon", "icon_state", "name", \
	                           "density" = 0, "opacity" = 0, "dir" = 2, "layer" = 0, \
	                           "pixel_x" = 0, "pixel_y" = 0, "pixel_z" = 0, "text", \
	                           "luminosity" = 0, "invisibility" = 0, "infra_luminosity" = 0, \
	                           "step_x" = 0, "step_y" = 0, "bound_x" = 0, "bound_y" = 0, \
	                           "bound_height" = 32, "bound_width" = 32 )

	var/projectName = "MDE_Map_Project"
	var/codeName = "_MDE_Objects"

	// Pass the configuration options to New()
	//
	// 'whitelist' sets which paths to include
	//    if null: only paths used in included maps are included
	//    if first path is null: append paths used in included maps to list
	//
	// 'blacklist' sets which paths to exclude
	//    if blank: only paths hidden from the object tree are excluded
	//    if first path is null: append paths hidden from the object tree to list
	//
	// 'projectName' sets the name of the generated map project
	//
	// 'includeRawIcons' determines if raw icon files are included or only the RSC
	//
	// 'additionalVariables' adds additional variables to be considered
	//
	New(list/whitelist = null, list/blacklist = null, projectName = null)
		ASSERT(!whitelist || istype(whitelist, /list))
		ASSERT(!blacklist || istype(blacklist, /list))

		src.projectName = projectName || src.projectName

		if(!whitelist) whitelist = list(null)
		if(!blacklist) blacklist = list(null)

		// Any specified paths auto-include their children
		var/count = whitelist.len
		for(var/i = 1 to count)
			var/path = whitelist[i]
			if(path && ispath(path))
				whitelist.Add(typesof(path) - path)

		// If the whitelist starts with null, include any paths used in maps
		if(count>0 && !whitelist[1])
			// Remove the null entry
			whitelist.Cut(1,2)
			// Get the default values
			whitelist.Add(GetIncludedMapPaths())

		count = blacklist.len
		for(var/i = 1 to count)
			var/path = blacklist[i]
			if(path && ispath(path))
				blacklist.Add(typesof(path) - path)

		// If the blacklist starts with null, include any paths hidden from the object tree
		if(count>0 && !blacklist[1])
			// Remove the null entry
			blacklist.Cut(1,2)
			// Get the default values
			blacklist.Add(GetHiddenTreePaths())

		var/includeList = DetermineIncludeList(whitelist, blacklist)

		GenerateProject(includeList)

	proc
		// Parses included map files for used paths (returns paths as text)
		GetIncludedMapPaths()

			// First find the DME
			var/list/files = flist("./")
			if(!files)
				world.log << "WARNING: MDE could not access file list, failed to determine paths used in maps"
				return list()

			var/dmeFile = ""
			for(dmeFile in files)
				if(findtext(dmeFile, ".dme", -4))
					break

			if(!dmeFile)
				world.log << "WARNING: MDE could not find DME file, failed to determine paths used in maps"
				return list()

			// Next parse the DME to get included map files
			var/dmeContents = file2text(dmeFile)
			if(!dmeContents)
				world.log << "WARNING: MDE could not access DME file '[dmeFile]', failed to determine paths used in maps"
				return list()

			var/list/result = list()

			// Check each line for a map file include
			var/list/dmeLines = LineSplit(dmeContents)
			for(var/line in dmeLines)
				var
					start = findtextEx(line, "#include \"")
					end = findtextEx(line, ".dmm\"") || findtextEx(line, ".dmp\"")
				if(start==1 && end==length(line)-4)
					// Parse the map file for any used paths
					var/list/paths = GetMapPaths(copytext(line, start+10, end+4))
					// Add the paths, while preventing duplicates
					for(var/path in paths) result[path] = 0

			return result

		// Parses particular map file for included paths (returns paths as text)
		GetMapPaths(dmmFile)
			// Map files start with an index of all tile data, including paths
			// There is a blank line separating the index from the actual map layout

			// Parse the DMM to get included paths
			var/dmmContents = file2text(dmmFile)
			if(!dmmContents)
				world.log << "WARNING: MDE could not access DMM file '[dmmFile]', failed to determine paths used"
				return list()

			// Check the map for paths
			var/list/result = list()
			var/list/paths = GatherPaths(dmmContents)
			for(var/path in paths) result[path] = 0

			return result

		// Parses INT file for hidden object tree paths (returns paths as text)
		GetHiddenTreePaths()

			// First find the INT
			var/list/files = flist("./")
			if(!files)
				world.log << "WARNING: MDE could not access file list, failed to determine hidden paths"
				return list()

			var/intFile = ""
			for(intFile in files)
				if(findtext(intFile, ".int", -4))
					break

			if(!intFile)
				world.log << "WARNING: MDE could not find INT file, failed to determine hidden paths"
				return list()

			// Next parse the INT to get hidden paths
			var/intContents = file2text(intFile)
			if(!intContents)
				world.log << "WARNING: MDE could not access INT file '[intFile]', failed to determine hidden paths"
				return list()

			var/list/result = list()

			// Check each line for the HIDDENPATHS token
			var/list/intLines = LineSplit(intContents)
			for(var/line in intLines)
				if(findtextEx(line, "HIDDENPATHS:") == 1)
					// Parse the line for any paths
					var/list/paths = GatherPaths(line)
					// Add the paths and their children
					for(var/path in paths)
						path = text2path(path)
						if(path)
							result.Add(typesof(path))

			return result

		// Combines whitelist and blacklist to determine final include list (returns paths as text)
		DetermineIncludeList(list/whitelist, list/blacklist)
			var/list/finalWhitelist = list()
			var/list/finalBlacklist = list()

			// All items in whitelist need to be converted to text
			// We also make sure that there are no duplicates by using associations
			for(var/path in whitelist)
				if(!istext(path)) path = "[path]"
				finalWhitelist[path] = 0

			// All items in blacklist need to be converted to text
			// We also make sure that there are no duplicates by using associations
			for(var/path in blacklist)
				if(!istext(path)) path = "[path]"
				finalBlacklist[path] = 0

			// Remove any black-listed paths from the whitelist
			finalWhitelist.Remove(finalBlacklist)

			return finalWhitelist

		GenerateProject(list/includeList)
			var/outputCode = GenerateWorld()

			// Generate code
			outputCode += GenerateGraph(includeList)

			// Overwrite existing generated map project
			if(fexists("[src.projectName]/"))
				fdel("[src.projectName]/")

			// Save object DM code file
			text2file(outputCode, "[src.projectName]/[src.codeName].dm")

			// Save project environment DME file
			text2file(GetEnvironmentString(), "[src.projectName]/[src.projectName].dme")

			// Find the RSC file
			var/list/files = flist("./")
			if(!files)
				world.log << "WARNING: MDE could not access file list, failed to include RSC file"
				return

			var/rscFile = ""
			for(rscFile in files)
				if(findtext(rscFile, ".rsc", -4))
					break

			if(!rscFile)
				world.log << "WARNING: MDE could not find RSC file, failed to include RSC file"
				return

			// Copy over RSC file
			fcopy(rscFile, "[src.projectName]/[src.projectName].rsc")

		GenerateWorld()
			. = {"
// Autogenerated map data from Map Data Extractor (DarkCampainger.MDE)
// Generated for [world.name] on [time2text(world.realtime, "Month DD, YYYY")]

world
	view = [ApplySyntax(world.view)]
	turf = [world.turf]
	area = [world.area]\n\n"}

		GenerateGraph(list/includeList)

			. = ""

			// "[type]" = PathNode
			var/list/lookup = list()

			// Create a column on the map to work out of
			world.maxx++
			var/createdMap = 0
			if(world.maxz == 0)
				world.maxy = 1
				world.maxz = 1
				createdMap = 1

			// Find a turf to create the atoms at
			var/turf/lab = locate(world.maxx, 1, 1)

			while(includeList.len)
				// Cheating, instead of sorting the list
				var/leaf = min(includeList)
				includeList -= leaf

				// For each path in includeList, go through its parents and then it
				var
					list/hierarchy = GetPathHierarchy(leaf)
					parent = null
					parameters = null
					tabbing = ""

				for(var/path in hierarchy)
					path = text2path(path)

					// See if this path is already generated
					parameters = lookup[path]

					// If not, we need to generate it
					if(!parameters)
						// Create it at our test tile
						var/datum/D = new path(lab)
						// Areas won't return the created instance?
						if(ispath(D)) D = locate(path)

						if(D)
							// Create a list of the instance's initial variables
							parameters = list("type" = path, "_mde_name_set" = 1, "_mde_text_set" = 1)
							for(var/v in savedVariables)
								parameters[v] = initial(D.vars[v])
							// Get the code for defining its attributes
							. += GetAttributes(parameters, parent, tabbing)
							// Store it in the lookup
							lookup[path] = parameters

					parent = parameters
					tabbing = "[tabbing]\t"

			// Remove the column
			world.maxx--
			if(createdMap)
				world.maxy = 0
				world.maxz = 0

			return .

		// Returns text string of code to set defined attributes
		GetAttributes(list/child, list/parent, tabbing="")

			. = "\n[tabbing][GetPathLeaf("[child["type"]]")]\n"

			for(var/attribute in savedVariables)
				if(!IsDefault(child, parent, attribute))
					. += "[tabbing]\t[attribute] = [ApplySyntax(child[attribute])]\n"

			return .

		IsDefault(list/child, list/parent, attribute)

			// Special case for names
			if(attribute == "name" && (!parent || !parent["_mde_name_set"]))
				if(GetPathName("[child["type"]]") == child["name"])
					child["_mde_name_set"] = 0
					return 1
				else
					child["_mde_name_set"] = 1
					return 0

			// Special case for text
			if(attribute == "text" && (!parent || !parent["_mde_text_set"]))
				if(child["name"] && copytext(child["name"], 1, 2) == child["text"])
					child["_mde_text_set"] = 0
					return 1
				else
					child["_mde_text_set"] = 1
					return 0

			if(parent)
				// See if the child's value deviates from the parent's
				return (child[attribute] == parent[attribute])

			else
				// Default is stored as associated value in savedVariables
				var/default = savedVariables[attribute]

				// Special case for bound_width/height
				if(attribute == "bound_height" || attribute == "bound_width")
					default = world.icon_size

				// Special case for layer
				if(attribute == "layer")
					if(ispath(child["type"], /area)) default = AREA_LAYER
					if(ispath(child["type"], /turf)) default = TURF_LAYER
					if(ispath(child["type"], /obj)) default = OBJ_LAYER
					if(ispath(child["type"], /mob)) default = MOB_LAYER

				// Special case for luminosity
				if(attribute == "luminosity" && ispath(child["type"],/area))
					default = 1

				// Special case for density
				if(attribute == "density" && ispath(child["type"],/mob))
					default = 1

				// See if the value differs from the default
				return (child[attribute] == default)

		GetEnvironmentString()
			return \
{"// DM Environment file for [projectName].dme.
// All manual changes should be made outside the BEGIN_ and END_ blocks.
// New source code should be placed in .dm files: choose File/New --> Code File.

// BEGIN_INTERNALS
// END_INTERNALS

// BEGIN_FILE_DIR
#define FILE_DIR .
// END_FILE_DIR

// BEGIN_PREFERENCES
// END_PREFERENCES

// BEGIN_INCLUDE
#include "[codeName].dm"
// END_INCLUDE"}


		///////////////////////
		// UTILITY FUNCTIONS //
		///////////////////////

		// Splits text into list of lines
		LineSplit(text)
			. = list()

			if(!text)
				return .

			var
				lineStart = 1
				lineEnd = 0

			while(TRUE)
				lineStart = lineEnd + 1
				if(text2ascii(text, lineStart) < 32) lineStart++

				lineEnd = findtextEx(text, ascii2text(13), lineStart) || findtextEx(text, ascii2text(10), lineStart)

				. += copytext(text, lineStart, lineEnd)

				if(lineEnd==0) break

			return .

		// Returns list of paths in text
		GatherPaths(text)
			. = list()

			var
				start = findtextEx(text, "/")
				end = start
				length = length(text)

			while(start)
				var/char
				do
					if(end == length) { end = 0; break }
					char = text2ascii(text, ++end)
				while((char>=47 && char<=57) || (char>=65 && char<=90) || (char>=97 && char<=122) || char==95)

				// Trim trailing slash
				if(text2ascii(text, end-1) == 47)
					. += copytext(text, start, end-1)
				else
					. += copytext(text, start, end)

				start = findtextEx(text, "/", end)
				end = start

			return .

		// Returns list of parent paths and path itself (paths returned as text)
		GetPathHierarchy(text)
			text = "[text]"

			var/result = list()

			var
				start = findtextEx(text, "/")
				end

			while(start)
				end = findtextEx(text, "/", start+1)
				result[copytext(text, 1, end)] = 0
				start = end

			return result

		// Return parent path based on text (path returned as text)
		GetPathParent(text)
			var
				indexA = 1
				indexB = 1

			while(indexB)
				indexA = indexB
				indexB = findtextEx(text, "/", indexB+1)

			return copytext(text, 1, indexA)

		// Returns the leaf of a path
		GetPathLeaf(text)
			var
				indexA = 1
				indexB = 1

			while(indexB)
				indexA = indexB
				indexB = findtextEx(text, "/", indexB+1)

			return copytext(text, indexA+1)

		// Returns the name of a path
		GetPathName(text)
			var/name = GetPathLeaf(text)

			var/index = 1
			while(TRUE)
				index = findtextEx(name, "_", index)
				if(index)
					name = copytext(name, 1, index) + " " + copytext(name, index+1, 0)
				else
					break

			return name

		// Apply proper to syntax to different types
		ApplySyntax(value)
			if(istext(value)) return "\"[value]\""
			else if(isfile(value)) return "'[value]'"
			else if(value == null) return "null"
			else return "[value]"